// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2023 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "nanogui/button.h"

#include "tools/side_menu.h"

#include "core/app_event.h"
#include "core/app_events_dispatcher.h"

#include "ui/nanogui/nanogui_side_menu.h"
#include "ui/nanogui/nanogui_main_menu.h"
#include "ui/nanogui/nanogui_about_pane.h"
#include "ui/nanogui/nanogui_granule_showcase.h"
#include "ui/nanogui/nanogui_extraction_dialog.h"
#include "ui/nanogui/nanogui_collections_control.h"
#include "ui/nanogui/nanogui_app_settings_editor.h"

#include "events/remote_pickle_export_request.h"
#include "events/remote_png_export_request.h"
#include "events/change_year_month_request.h"
#include "events/change_year_request.h"
#include "events/side_menu_resize_request.h"
#include "events/display_catalogue_request.h"
#include "events/granule_info_ready_for_display_event.h"
#include "events/display_extraction_info_request.h"
#include "events/collection_removal_request.h"
#include "events/app_settings_changed_event.h"


using namespace SEAScope;

SideMenu::SideMenu(AppEventsDispatcher& appEventsDispatcher, nanogui::Widget* parent, bool processingToolsAvailable)
: _appEventsDispatcher(appEventsDispatcher),
  _ui(nullptr),
  _mainMenu(nullptr),
  _catalogueButton(nullptr),
  _granuleInfoButton(nullptr),
  _extractedDataButton(nullptr),
  _settingsButton(nullptr),
  _aboutButton(nullptr)
{
	this->_ui = new NanoguiSideMenu(parent);
	this->_ui->setFixedWidth(320);
	this->_ui->setFixedHeight(parent->height() - 136 /* timeline height */);
	this->_ui->anchor->validCorners.push_back(NanoguiCorner::TOP_LEFT);
	this->_ui->anchor->validCorners.push_back(NanoguiCorner::TOP_RIGHT);
	this->_ui->content = new NanoguiMainMenu(this->_ui);
	this->_ui->content->setPosition(Eigen::Vector2i(0,0));
	this->_ui->anchor->setCorner(NanoguiCorner::TOP_RIGHT);
	this->_ui->setSize(this->_ui->fixedSize());
	this->_ui->setVisible(false);

	this->_mainMenu = dynamic_cast<NanoguiMainMenu*>(this->_ui->content);
	this->_mainMenu->setFixedWidth(320);
    this->_mainMenu->update();

	// Hide processing options in the extraction dialog if the processor is
	// not enabled
    if
        (true == processingToolsAvailable)
    {
	    this->_mainMenu->getExtractionDialog()->showProcessingOptions();
    }
    else
    {
	    this->_mainMenu->getExtractionDialog()->hideProcessingOptions();
    }

	// Sidemenu header
	this->_ui->header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
													    nanogui::Alignment::Middle,
													    2, 3));
	this->_catalogueButton = new nanogui::Button(this->_ui->header, "",
                                                 ENTYPO_ICON_DATABASE);
    this->_catalogueButton->setFlags(nanogui::Button::RadioButton);
    this->_catalogueButton->setTooltip("Data catalogue");

    this->_granuleInfoButton = new nanogui::Button(this->_ui->header, "",
                                                   ENTYPO_ICON_INFO);
    this->_granuleInfoButton->setFlags(nanogui::Button::RadioButton);
    this->_granuleInfoButton->setTooltip("Information about selected granule");

    this->_extractedDataButton = new nanogui::Button(this->_ui->header, "",
                                                     ENTYPO_ICON_PUBLISH);
    this->_extractedDataButton->setFlags(nanogui::Button::RadioButton);
    this->_extractedDataButton->setTooltip("Information about extracted data");

    this->_settingsButton = new nanogui::Button(this->_ui->header, "",
                                                ENTYPO_ICON_COG);
    this->_settingsButton->setFlags(nanogui::Button::RadioButton);
    this->_settingsButton->setTooltip("SEAScope settings");

	this->_aboutButton = new nanogui::Button(this->_ui->header, "",
                                            ENTYPO_ICON_HELP);
    this->_aboutButton->setFlags(nanogui::Button::RadioButton);
    this->_aboutButton->setTooltip("Information about SEAScope");

    this->_tabButtons.push_back(this->_catalogueButton);
    this->_tabButtons.push_back(this->_granuleInfoButton);
    this->_tabButtons.push_back(this->_extractedDataButton);
    this->_tabButtons.push_back(this->_settingsButton);
    this->_tabButtons.push_back(this->_aboutButton);
    this->_catalogueButton->setButtonGroup(this->_tabButtons);
    this->_granuleInfoButton->setButtonGroup(this->_tabButtons);
    this->_extractedDataButton->setButtonGroup(this->_tabButtons);
    this->_settingsButton->setButtonGroup(this->_tabButtons);
    this->_aboutButton->setButtonGroup(this->_tabButtons);

	this->_mainMenu->getCollectionsControl()->setListChangedCallback([this] () {
        this->_mainMenu->update();
		this->_ui->reflowNeeded = true;
	});
    this->_mainMenu->getCollectionsControl()->setCollectionRemovedCallback([this](const CollectionId& colId) {
        CollectionRemovalRequest event(colId);
        this->_appEventsDispatcher.dispatch<CollectionRemovalRequest>(event);
    });

	this->_mainMenu->getGranuleInfo()->monthCallback = [this] (std::uint16_t year, std::uint8_t month) {
        ChangeYearMonthRequest event(year, month);
        this->_appEventsDispatcher.dispatch<ChangeYearMonthRequest>(event);
		//app.engine.timeline.view->setMonth(year, month);
	};
	this->_mainMenu->getGranuleInfo()->yearCallback = [this] (std::uint16_t year) {
        ChangeYearRequest event(year);
        this->_appEventsDispatcher.dispatch<ChangeYearRequest>(event);
		// app.engine.timeline.view->setYear(year);
	};

	this->_mainMenu->getExtractionDialog()->pickleCallback = [this] () {
        RemotePickleExportRequest event;
        this->_appEventsDispatcher.dispatch<RemotePickleExportRequest>(event);
	};
	this->_mainMenu->getExtractionDialog()->pngCallback = [this] () {
        RemotePNGExportRequest event;
        this->_appEventsDispatcher.dispatch<RemotePNGExportRequest>(event);
	};
    this->_mainMenu->getSettingsPane()->setChangeCallback([this] (const Settings& settings) {
        AppSettingsChangedEvent event{settings};
        this->_appEventsDispatcher.dispatch<AppSettingsChangedEvent>(event);
    });

	this->_catalogueButton->setChangeCallback([this](bool pushed) {
		this->_mainMenu->getCollectionsControl()->setVisible(pushed);
		this->_catalogueButton->setPushed(pushed);
		this->_mainMenu->update();
	});

    this->_granuleInfoButton->setChangeCallback([this](bool pushed) {
        this->_mainMenu->getGranuleInfo()->setVisible(pushed);
        this->_granuleInfoButton->setPushed(pushed);
        this->_mainMenu->update();
    });

    this->_extractedDataButton->setChangeCallback([this](bool pushed) {
        this->_mainMenu->getExtractionDialog()->setVisible(pushed);
        this->_extractedDataButton->setPushed(pushed);
        this->_mainMenu->update();
    });

    this->_settingsButton->setChangeCallback([this](bool pushed) {
        this->_mainMenu->getSettingsPane()->setVisible(pushed);
        this->_settingsButton->setPushed(pushed);
        this->_mainMenu->update();
    });

	this->_aboutButton->setChangeCallback([this](bool pushed) {
        this->_mainMenu->getAboutPane()->setVisible(pushed);
        this->_aboutButton->setPushed(pushed);
        this->_mainMenu->update();
	});

    this->_appEventsDispatcher.subscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.subscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.subscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.subscribe<DisplayExtractionInfoRequest>(this);
}

SideMenu::~SideMenu()
{
    this->_appEventsDispatcher.unsubscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayExtractionInfoRequest>(this);

    this->_tabButtons.clear();
}

/**
 *
 */
bool SideMenu::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SideMenuResizeRequest)) == eventType)
    {
        int width = static_cast<const SideMenuResizeRequest&>(event).getWidth();
        int height = static_cast<const SideMenuResizeRequest&>(event).getHeight();

		this->_ui->anchor->update();
        this->_ui->setFixedSize(Eigen::Vector2i(width, height));
        this->_ui->setSize(Eigen::Vector2i(width, height));
        this->_mainMenu->update();
        this->_ui->reflowNeeded = true;

        return true;
    }
    else if
        (std::type_index(typeid(GranuleInfoReadyForDisplayEvent)) == eventType)
    {
        const GranuleInfo& info = static_cast<const GranuleInfoReadyForDisplayEvent&>(event).getGranuleInfo();
		this->_mainMenu->getGranuleInfo()->displayInfo(info);

        if
            (false == this->_ui->visible())
        {
		    this->_ui->setVisible(true);
        }

        if
            (false == this->_mainMenu->getGranuleInfo()->visible())
        {
            this->_aboutButton->setPushed(false);
            this->_mainMenu->getAboutPane()->setVisible(false);

            this->_settingsButton->setPushed(false);
            this->_mainMenu->getSettingsPane()->setVisible(false);

            this->_catalogueButton->setPushed(false);
            this->_mainMenu->getCollectionsControl()->setVisible(false);

            this->_extractedDataButton->setPushed(false);
            this->_mainMenu->getExtractionDialog()->setVisible(false);

            this->_granuleInfoButton->setPushed(true);
            this->_mainMenu->getGranuleInfo()->setVisible(true);
        }

        this->_mainMenu->update();

        return true;
    }
    else if
        (std::type_index(typeid(DisplayCatalogueRequest)) == eventType)
    {
		if
			(false == this->_ui->visible())
		{
			this->_ui->setVisible(true);
		}

		if
			(false == this->_mainMenu->getCollectionsControl()->visible())
		{
            this->_aboutButton->setPushed(false);
            this->_mainMenu->getAboutPane()->setVisible(false);

            this->_settingsButton->setPushed(false);
            this->_mainMenu->getSettingsPane()->setVisible(false);

            this->_catalogueButton->setPushed(true);
            this->_mainMenu->getCollectionsControl()->setVisible(true);

            this->_extractedDataButton->setPushed(false);
            this->_mainMenu->getExtractionDialog()->setVisible(false);

            this->_granuleInfoButton->setPushed(false);
            this->_mainMenu->getGranuleInfo()->setVisible(false);
		}

        this->_mainMenu->update();

        return true;
    }
    else if
        (std::type_index(typeid(DisplayExtractionInfoRequest)) == eventType)
    {
        bool success(false);

        const std::map<RenderableId, Extraction>& extractionData = static_cast<const DisplayExtractionInfoRequest&>(event).getExtractionData();
        const std::map<RenderableId, std::string>& labels = static_cast<const DisplayExtractionInfoRequest&>(event).getLabels();
        this->_mainMenu->getExtractionDialog()->setExtraction(extractionData, labels, success);

		if
			(false == this->_ui->visible())
		{
			this->_ui->setVisible(true);
		}

        if
            (false == this->_mainMenu->getExtractionDialog()->visible())
        {
            this->_aboutButton->setPushed(false);
            this->_mainMenu->getAboutPane()->setVisible(false);

            this->_settingsButton->setPushed(false);
            this->_mainMenu->getSettingsPane()->setVisible(false);

            this->_catalogueButton->setPushed(false);
            this->_mainMenu->getCollectionsControl()->setVisible(false);

            this->_extractedDataButton->setPushed(true);
            this->_mainMenu->getExtractionDialog()->setVisible(true);

            this->_granuleInfoButton->setPushed(false);
            this->_mainMenu->getGranuleInfo()->setVisible(false);
        }

        this->_mainMenu->update();

        return true;
    }

    return false;
}

NanoguiMainMenu* SideMenu::getMainMenu() const
{
    return this->_mainMenu;
}

/*
	this->_ui->anchor->updateCallback(0, 0, 0, 0); //Hack required with some window managers (the other workaround is to resize the window)
*/
