// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2023 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cmath>
#include <cstddef>
#include <sstream>
#include <iomanip>
#include <algorithm>

#include "core/log.h"
#include "data/image.h"
#include "local/idf_reader.h"
#include "local/netcdf_utils.h"
#include "vendors/_netcdf.h"
#include "local/reader_cfg.h"
#include "core/granule_metadata.h"

using namespace SEAScope;



/**
 *
 */
bool IDFReader::extractGranuleInfo(const std::string& path, std::vector<GranuleMetadata>& granules, const ReaderCfg&) const
{
    bool ok {false}; //
    std::time_t tmpTimestamp {};    //
    netCDF::NcVar time; //
	netCDF::NcVar lat;	//
	netCDF::NcVar lon;	//
	netCDF::NcVar x;	//
	netCDF::NcVar y;	//
	netCDF::NcGroupAtt attr;	//
	std::size_t gcpCount;	//
	std::size_t i;	//
	std::size_t j;	//
    double timeValue;   //
    double* times;  //
	float lon0;	//
	float* lats;	//
	float* lons;	//
	std::int32_t* xs;	//
	std::int32_t* ys;	//
	float resolution;	//
	std::string resolutionUnits;	//
	std::string timeCoverageStartStr;	//
	std::string timeCoverageStopStr;	//
    IDFDescriptor descriptor;   //

    DataModel dataModel{DataModel::NONE};   //
    std::string dataId; //
    timestamp_t timeCoverageStart {0u};  //
    timestamp_t timeCoverageEnd {0u};    //

	Log::log(LogLevel::Debug, "IDFReader", "extractGranuleInfo({})", path);
    netCDF::NcFile fileHandler(path, netCDF::NcFile::read);

	if
		(true == fileHandler.isNull())
	{
		Log::log(LogLevel::Error, "IDFReader", "Could not open file {}", path);
		return false;
	}

    granules.emplace_back();
    GranuleMetadata& info = granules.back();

    descriptor.uri = path;

	/* -------------------- Extract mandatory attributes --------------------*/

	// idf_granule_id
    attr = fileHandler.getAtt("idf_granule_id");
	if
		(true == attr.isNull())
	{
		Log::log(LogLevel::Error, "IDFReader"
				, "Global attribute idf_granule_id missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
    attr.getValues(dataId);

    // time_coverage_start
	attr = fileHandler.getAtt("time_coverage_start");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_start missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
	attr.getValues(timeCoverageStartStr);
    ok = SEAScope::str2timestamp(timeCoverageStartStr, tmpTimestamp);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "IDFReader", "Failed to parse time_coverage_start");
        fileHandler.close();
        return false;
    }
    timeCoverageStart = static_cast<timestamp_t>(tmpTimestamp);
	//Log::log(LogLevel::Debug, "IDFReader", "Time coverage start: {} {}", timeCoverageStartStr, info.start);

    // time_coverage_end
	attr = fileHandler.getAtt("time_coverage_end");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_end missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
	attr.getValues(timeCoverageStopStr);
    ok = SEAScope::str2timestamp(timeCoverageStopStr, tmpTimestamp);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "IDFReader", "Failed to parse time_coverage_end");
        fileHandler.close();
        return false;
    }
	timeCoverageEnd = static_cast<timestamp_t>(tmpTimestamp);
	//Log::log(LogLevel::Debug, "IDFReader", "Time coverage stop: {} {}", timeCoverageStopStr, info.stop);

	// resolution
    attr = fileHandler.getAtt("idf_spatial_resolution");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute idf_spatial_resolution missing from {}"
				, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
    attr.getValues(&resolution);

	// resolution units
	attr = fileHandler.getAtt("idf_spatial_resolution_units");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
			, "Global attribute idf_spatial_resolution_units missing from {}"
			, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
	attr.getValues(resolutionUnits);

	if
		("m" != resolutionUnits)
	{
		/* TODO: convert resolution in meters */
	}
	descriptor.resolution = resolution;

	// subsampling factor
	attr = fileHandler.getAtt("idf_subsampling_factor");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute idf_subsampling_factor missing from {}"
				, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
	attr.getValues(&descriptor.subsamplingFactor);

	/* ------------------- Extract optional attributes ----------------------*/
    IDFOptionalAttributes optionalAttrs;
    SEAScope::extractOptionalAttributes(fileHandler, optionalAttrs);


	/* Find out which model is used */
	std::vector<std::string> modelRowCell;	//
	modelRowCell.push_back("time");
	modelRowCell.push_back("cell");
	modelRowCell.push_back("row");

	std::vector<std::string> modelLatLon;	//
	modelLatLon.push_back("time");
	modelLatLon.push_back("lat");
	modelLatLon.push_back("lon");

	std::vector<std::string> modelYX;	//
	modelYX.push_back("time");
	modelYX.push_back("y");
	modelYX.push_back("x");

	std::vector<std::string> modelNjNi;	//
	modelNjNi.push_back("time");
	modelNjNi.push_back("nj");
	modelNjNi.push_back("ni");

	std::vector<std::string> modelStation;	//
	modelStation.push_back("time");
	modelStation.push_back("station");

	std::vector<std::string> modelTraj;	//
	modelTraj.push_back("time");

	auto dims = fileHandler.getDims();
	for
		(const auto& it : dims)
	{
		auto foundLatLon = std::find(modelLatLon.begin(), modelLatLon.end(), it.first);
		if
			(modelLatLon.end() != foundLatLon)
		{
			modelLatLon.erase(foundLatLon);
		}

		auto foundRowCell = std::find(modelRowCell.begin(), modelRowCell.end(), it.first);
		if
			(modelRowCell.end() != foundRowCell)
		{
			modelRowCell.erase(foundRowCell);
		}

		auto foundYX = std::find(modelYX.begin(), modelYX.end(), it.first);
		if
			(modelYX.end() != foundYX)
		{
			modelYX.erase(foundYX);
		}

		auto foundNjNi = std::find(modelNjNi.begin(), modelNjNi.end(), it.first);
		if
			(modelNjNi.end() != foundNjNi)
		{
			modelNjNi.erase(foundNjNi);
		}

		auto foundStation = std::find(modelStation.begin(), modelStation.end(), it.first);
		if
			(modelStation.end() != foundStation)
		{
			modelStation.erase(foundStation);
		}

		auto foundTraj = std::find(modelTraj.begin(), modelTraj.end(), it.first);
		if
			(modelTraj.end() != foundTraj)
		{
			modelTraj.erase(foundTraj);
		}
		Log::log(LogLevel::Debug, "IDFReader", "Dimension {}: {}", it.first, it.second.getSize());
	}

	gcpCount = 0;
	if
		(0 >= modelLatLon.size())
	{
		dataModel = DataModel::LAT_LON;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from lat/lon model");
		descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
        time = fileHandler.getVar("time");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_lat_gcp");
		x = fileHandler.getVar("index_lon_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lons = new float[descriptor.shape.xArity];
		lats = new float[descriptor.shape.yArity];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

        time.getVar(&timeValue);
		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

        descriptor.shape.gcpsTime.reserve(gcpCount);
        descriptor.shape.gcpsLat.reserve(gcpCount);
        descriptor.shape.gcpsLon.reserve(gcpCount);
        descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
        descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
                descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
                descriptor.shape.gcpsLat.push_back(lats[j]);
                descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0)));
                descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
                descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
				//Log::log(LogLevel::Error, "IDFReader", "longitude: {} -> {}", lons[i], std::fmod(lons[i] -lon0, 360.0) + lon0);
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelRowCell.size())
	{
		dataModel = DataModel::ROW_CELL;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from row/cell model");
		descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
        time = fileHandler.getVar("time");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_row_gcp");
		x = fileHandler.getVar("index_cell_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

        time.getVar(&timeValue);
		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

        descriptor.shape.gcpsTime.reserve(gcpCount);
        descriptor.shape.gcpsLat.reserve(gcpCount);
        descriptor.shape.gcpsLon.reserve(gcpCount);
        descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
        descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
                descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
                descriptor.shape.gcpsLat.push_back(lats[j * descriptor.shape.xArity + i]);
                descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0)));
                descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
                descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelYX.size())
	{
		dataModel = DataModel::Y_X;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from y/x model");
		descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
        time = fileHandler.getVar("time");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_y_gcp");
		x = fileHandler.getVar("index_x_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

        time.getVar(&timeValue);
		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

        descriptor.shape.gcpsTime.reserve(gcpCount);
        descriptor.shape.gcpsLat.reserve(gcpCount);
        descriptor.shape.gcpsLon.reserve(gcpCount);
        descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
        descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
                descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * timeValue) /*same time for all GCPs*/);
                descriptor.shape.gcpsLat.push_back(lats[j * descriptor.shape.xArity + i]);
                descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0)));
                descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
                descriptor.shape.gcpsMajorAxisIndex.push_back(static_cast<uint32_t>(ys[j]));
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelNjNi.size())
	{
		dataModel = DataModel::NJ_NI;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from nj/ni model");
		descriptor.shape.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni").getSize());
		y = fileHandler.getVar("index_nj_gcp");
		x = fileHandler.getVar("index_ni_gcp");
	}
	else if
		(0 >= modelStation.size())
	{
		dataModel = DataModel::TIME_STATION;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from station model");
		descriptor.shape.yArity = 1;
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("station_gcp").getSize());
		descriptor.yArity = 1;
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
		x = fileHandler.getVar("index_station_gcp");
	}
	else if
		(0 >= modelTraj.size())
	{
		dataModel = DataModel::TIME;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from trajectory model");
		descriptor.shape.yArity = 1;
		descriptor.shape.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time_gcp").getSize());
		descriptor.yArity = 1;
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
        time = fileHandler.getVar("time");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		x = fileHandler.getVar("index_time_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get {} values", descriptor.shape.xArity);
		gcpCount = descriptor.shape.xArity;
        times = new double[gcpCount];
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];

        time.getVar(times);
		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

        descriptor.shape.gcpsTime.reserve(gcpCount);
        descriptor.shape.gcpsLat.reserve(gcpCount);
        descriptor.shape.gcpsLon.reserve(gcpCount);
        descriptor.shape.gcpsMinorAxisIndex.reserve(gcpCount);
        descriptor.shape.gcpsMajorAxisIndex.reserve(gcpCount);
        i = 0;
        for
            (; i < descriptor.shape.xArity; ++i)
        {
            descriptor.shape.gcpsTime.push_back(static_cast<std::uint64_t>(1000 * times[i]));
            descriptor.shape.gcpsLat.push_back(lats[i]);
            descriptor.shape.gcpsLon.push_back(static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0)));
            descriptor.shape.gcpsMinorAxisIndex.push_back(static_cast<uint32_t>(xs[i]));
            descriptor.shape.gcpsMajorAxisIndex.push_back(0u);
        }

        delete [] times;
        times = nullptr;
		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
	}

	Log::log(LogLevel::Debug, "IDFReader", "Rebuild GCPs list");

    info.build(0u, 0u, 0u,
               dataModel, timeCoverageStart, timeCoverageEnd,
               optionalAttrs.hasLatMin, optionalAttrs.latMin,
               optionalAttrs.hasLatMax, optionalAttrs.latMax,
               optionalAttrs.hasLonMin, optionalAttrs.lonMin,
               optionalAttrs.hasLonMax, optionalAttrs.lonMax,
               dataId,
               optionalAttrs.title,
               optionalAttrs.institution,
               optionalAttrs.comment,
               optionalAttrs.fileId,
               optionalAttrs.productVersion,
               optionalAttrs.creatorEmail,
               optionalAttrs.stationId,
               optionalAttrs.platform,
               optionalAttrs.sensor,
               descriptor);

	Log::log( LogLevel::Debug, "IDFReader"
			, "GranuleInfo extracted!");

    times = nullptr;
	lons = nullptr;
	lats = nullptr;
	xs = nullptr;
	ys = nullptr;

	fileHandler.close();

	return true;
};

/**
 *
 */
bool IDFReader::extractGranuleData(const std::string& path, const std::set<std::string>& fields, DataModel dataModel, GranuleData& data) const
{
	netCDF::NcVar var;	//
	netCDF::NcVarAtt offsetAtt;	//
	netCDF::NcVarAtt scaleAtt;	//
	netCDF::NcVarAtt fillValueAtt;	//
	float offset;	//
	float scaleFactor;	//
	float fillValue;	//
	std::uint8_t* ubyteResult;	//
    float* floatResult; //

	bool ok;	//

	ubyteResult = nullptr;
    floatResult = nullptr;

    netCDF::NcFile fileHandler(path.c_str(), netCDF::NcFile::read);
    if
        (true == fileHandler.isNull())
    {
        Log::log( LogLevel::Error, "IDFReader", "Could not open netCDF file {}"
                , path);
        return false;
    }

	for
		(const auto& field : fields)
	{
		Log::log(LogLevel::Debug, "IDFReader", "Extracting field: {}", field);
		var = fileHandler.getVar(field);

        if
            (true == var.isNull())
        {
            Log::log(LogLevel::Error, "IDFReader", "Unknown field: {}", field);
            return false;
        }

        data[field];
        DataInfo& info = data[field].info;

		if
			(DataModel::LAT_LON == dataModel)
		{
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon").getSize());
			info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat").getSize());
		}
		else if
			(DataModel::Y_X == dataModel)
		{
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x").getSize());
			info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y").getSize());
		}
		else if
			(DataModel::NJ_NI == dataModel)
		{
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni").getSize());
			info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj").getSize());
		}
		else if
			(DataModel::ROW_CELL == dataModel)
		{
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell").getSize());
			info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row").getSize());
		}
		else if
			(DataModel::TIME_STATION == dataModel)
		{
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
			info.yArity = static_cast<std::uint32_t>(fileHandler.getDim("station").getSize());
		}
		else if
			(DataModel::TIME == dataModel)
		{
            Log::log(LogLevel::Debug, "IDFReader", "extract data from trajectory");
			info.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
			info.yArity = 1u;
		}
        else
        {
            Log::log(LogLevel::Error, "IDFReader", "Unknown data model: {}",
                     static_cast<std::uint8_t>(dataModel));
            return false;
        }

		if
			(  "time" == field
			&& DataModel::TIME == dataModel)
		{
			// Time is stored as double precision but SEAScope only handles
			// floats.
			double* timeBuffer = new double[info.xArity];
			double minTime;
			double maxTime;
			double timeExtent;
			std::int64_t n;
			Log::log(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);

			var.getVar(timeBuffer);

			minTime = std::numeric_limits<double>::max();
			maxTime = std::numeric_limits<double>::min();
			n = static_cast<std::int64_t>(info.xArity);
			while
				(--n > -1)
			{
				if
					(timeBuffer[n] > maxTime)
				{
					maxTime = timeBuffer[n];
				}
				if
					(timeBuffer[n] < minTime)
				{
					minTime = timeBuffer[n];
				}
			}
			Log::log(LogLevel::Debug, "IDFReader", "Time range: [{}, {}]", minTime, maxTime);

			timeExtent = maxTime - minTime;
			Log::log(LogLevel::Debug, "IDFReader", "Time extent: {}", timeExtent / 86400);
			if
				(timeExtent > double(std::numeric_limits<float>::max()))
			{
				Log::log(LogLevel::Error, "IDFReader", "Overflow: the time range is too wide to be stored as a float value");
				fileHandler.close();
				return false;
			}
			offset = float(minTime);
			scaleFactor = 1.f;

			info.channels = 1;
			info.dataType = DataType::float32;
			info.offsets.push_back(offset);
			info.scaleFactors.push_back(scaleFactor);
			info.fillValues.push_back(-1.f);

			data[field].buffer.allocate(info.xArity, info.yArity, info.channels,
										info.dataType, nullptr);
			floatResult = static_cast<float*>(data[field].buffer.data());
			n = static_cast<std::int64_t>(info.xArity);
			while
				(--n > -1)
			{
				float tmpValue = static_cast<float>(timeBuffer[n] - static_cast<double>(offset)) * scaleFactor;
				Log::log(LogLevel::Debug, "IDFReader", "n = {}, tmpValue = {}" , n, tmpValue);
				floatResult[n] = tmpValue;
			}

			n = static_cast<std::int64_t>(info.xArity);
			for
				(int i = 0; i < n; ++i)
			{
				Log::log(LogLevel::Debug, "IDFReader", "result: {} + ({} * {})", offset, floatResult[i], scaleFactor);
			}

			delete[] timeBuffer;
		}
		else
		{
			info.channels = 1;

			/* Get offset */
			offsetAtt = SEAScope::readAttribute(var, "add_offset", ok);
			offset = 0.f;
			if
				(true == ok)
			{
				offsetAtt.getValues(&offset);
			}
			info.offsets.push_back(offset);

			/* Get scale */
			scaleAtt = SEAScope::readAttribute(var, "scale_factor", ok);
			scaleFactor = 1.f;
			if
				(true == ok)
			{
				scaleAtt.getValues(&scaleFactor);
			}
			info.scaleFactors.push_back(scaleFactor);

			/* Get fill value */
			fillValueAtt = SEAScope::readAttribute(var, "_FillValue", ok);
			if
				(true == ok)
			{
				fillValueAtt.getValues(&fillValue);
			}
			else
			{
				Log::log(LogLevel::Error, "IDFReader", "Cannot find fill value for {}", field);
				fileHandler.close();
				return false;
			}
			info.fillValues.push_back(fillValue);

            netCDF::NcType::ncType dataType = var.getType().getTypeClass();
            if
                (netCDF::NcType::ncType::nc_FLOAT == dataType)
            {
			    info.dataType = DataType::float32;
                data[field].buffer.allocate(info.xArity, info.yArity,
                                            info.channels, info.dataType,
                                            nullptr);
			    floatResult = static_cast<float*>(data[field].buffer.data());
			    var.getVar(floatResult);

            }
            else if
                (netCDF::NcType::ncType::nc_UBYTE == dataType)
            {
			    info.dataType = DataType::unsigned_byte;
                data[field].buffer.allocate(info.xArity, info.yArity,
                                            info.channels, info.dataType,
                                            nullptr);
			    ubyteResult = static_cast<std::uint8_t*>(data[field].buffer.data());
			    var.getVar(ubyteResult);
            }
            else
            {
                Log::log(LogLevel::Error, "IDFReader",
                         "Data type not supported for field {}: {} (only ubyte and float32 are accepted)",
                          field, int(dataType));
                return false;
            }

			Log::log(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);
		}
	}

	fileHandler.close();
	ubyteResult = nullptr;
    floatResult = nullptr;
	return true;
};
